#!/usr/bin/perl

#******************************************************************************
# Programa TCAalign
# Alinha as sentencas dos textos paralelos passados como parametro.
# Criterios de alinhamento:
# Cognatos: LCSR (threshold=0.65) ou Coeficiente de Dice (threshold=0.65)
# Candidatas a nomes proprios: palavras com inicial maiuscula
# Caracteres especiais: ?,!,$,@,%
# Lista de palavras ancoras (opcional): arquivo como parametro
# Entrada: Dois arquivos com cada sentenca em uma linha OU etiquetados com 
# delimitadores de texto (text), paragrafo (p) e sentenca(s) (usar opcao -tg)
# Saida: Dois arquivos com extensao .al 
#
# OBS.: Ha uma opcao de gerar apenas alinhamentos 1:1 (-o)
#******************************************************************************

use FileHandle;
use locale;
use strict;
use Getopt::Long;
use Pod::Usage;

my ($arqfonte,$arqalvo,$lpa,$help,$verbose,$um_para_um,$corpus,%ancfonte,%ancalvo);

# Parametros
my $lowind = 0.20; 
my $highind = 0.40;
my $toohigh = 0.50;
my $limite = 0.65;	
my $janela = 5;
my $metrica = 'dice';
my $etiquetado = $um_para_um = 0;
$corpus = $lpa = '';

GetOptions( 'source|s=s' 			=> \$arqfonte,
				'target|t=s' 			=> \$arqalvo,
				'corpus|c=s' 			=> \$corpus,
				'lpa|l=s' 				=> \$lpa,
				'one_to_one|o' 		=> \$um_para_um,
				'cog_metric|m' 		=> \$metrica,
				'cog_threshold|th=f' => \$limite,
				'tagged|tg'				=> \$etiquetado,
				'help|h|?'				=> \$help,
				'verbose|v'				=> \$verbose
			) || pod2usage(2);
	pod2usage(2) if $help;
	pod2usage(2) unless ($arqfonte && $arqalvo) || ($corpus);
	if (($metrica ne 'lcsr') && ($metrica ne 'dice')) { die "Choose one of these cognate metrics: lcsr or dice\n"; }

if ($lpa ne '') { le_LPA($lpa,%ancfonte,%ancalvo); }
else { %ancfonte = %ancalvo = (); }

if ($corpus ne '') {
	open(ARQ,$corpus) or die "It is not possible to open corpus file\n";
	my @arqs = <ARQ>;
	close ARQ;
	while ($#arqs >= 0) {
		($arqfonte,$arqalvo) = split(/\|/,shift(@arqs));
		$arqalvo =~ s/\n//g;
		if (alinha($arqfonte,$arqalvo,\%ancfonte,\%ancalvo) == 0) { print "Alignment Finished\n"; exit; }
	}
}
else { alinha($arqfonte,$arqalvo); }
print "\n";

# Retorna -1 se $a for menor que $b e 1 caso contrario
sub compara {
	my($a,$b) = @_;
	my($sa,$sb,$pa,$pb);

	$a =~ /[^\.]+\.(\d+)\.s(\d+)/;
	$pa = int($1);
	$sa = int($2);
	$b =~ /[^\.]+\.(\d+)\.s(\d+)/;
	$pb = int($1);
	$sb = int($2);
	if ($pa < $pb) { return -1; }
	elsif ($pa == $pb) {
		if ($sa < $sb) { return -1; }
	}
	return 1;
}

sub alinha{
	my($arqfonte,$arqalvo,$htrans,$vtrans) = @_;
	my($saida) = 1;
	my (%sfonte,%salvo,@lenf,@lena,@listaf,@listaa,@matrix,$alinhado,@auxfonte,@auxalvo);
	my($l,$c,$contf,$conta,$qtdf,$qtda,$melhor,@idsf,@idsa,@alf,@ala,$cabf,$caba,$i,@aux);
	my $idali = 1; # indice do alinhamento quando a entrada nao esta etiquetada

	%sfonte = %salvo = @lenf = @lena = @listaf = @listaa = @matrix = @auxfonte = @auxalvo = ();
	
	print "\nAligning $arqfonte and $arqalvo ...\n";  
	$alinhado = nome($arqfonte).'.al';
	open(fout,">$alinhado") or die "It is not possible to open file $alinhado\n";
	close fout;
	$alinhado = nome($arqalvo).'.al';
	open(fout,">$alinhado") or die "It is not possible to open file $alinhado\n";
	close fout;
	
	if ($etiquetado) {
		$cabf = le_sentencas_etiquetadas($arqfonte,\%sfonte);
		$caba = le_sentencas_etiquetadas($arqalvo,\%salvo);
		@idsf = sort {compara($a,$b)} keys %sfonte;
		@idsa = sort {compara($a,$b)} keys %salvo;
	}
	else {
		le_sentencas_cruas($arqfonte,\%sfonte);
		le_sentencas_cruas($arqalvo,\%salvo);
		@idsf = sort {int($a) <=> int($b)} keys %sfonte;
		@idsa = sort {int($a) <=> int($b)} keys %salvo;
	}	
	
	$contf = $conta = 0;
	while (($#idsf >= 0) && ($#idsa >= 0)) { 
		if ($um_para_um) { $qtdf = $qtda = 1; }
		else {
			# Busca sentencas a alinhar
			while (($contf < $janela) && ($contf <= $#idsf)) { 
				listas($sfonte{$idsf[$contf]},$contf,$htrans,\@listaf);
				push(@lenf,length($sfonte{$idsf[$contf]}));
				$contf++;
			}
			while (($conta < $janela) && ($conta <= $#idsa)) { 
				listas($salvo{$idsa[$conta]},$conta,$vtrans,\@listaa);
				push(@lena,length($salvo{$idsa[$conta]}));
				$conta++;
			}

			if ($#matrix < 0) { 
				$qtdf = $contf; 
				$qtda = $conta; 			
			}
			
			if ($verbose){
				imprime_listas(\@listaf);		
				imprime_listas(\@listaa);
			}
    			#prepara a matriz para receber os valores das combinacoes de sentencas
			zera_matrix($contf-$qtdf,$contf-1,$conta-$qtda,$conta-1,\@matrix);
			# Completa a matriz
			completa_matrix($contf-$qtdf,$contf-1,$conta-$qtda,$conta-1,\@listaf,\@listaa,$htrans,\%sfonte,\@idsf,\%salvo,\@idsa,\@matrix);

			if ($verbose) { imprime_matrix(\@lenf,\@lena,$conta,$contf,\@matrix); }
		
			$melhor = combinacoes($contf,$conta,\@lenf,\@lena,\@matrix);	
			@aux = split(/\n/,$melhor);	
			($qtdf,$qtda) = split(/:/,$aux[0]);
			if (($qtdf == 0) && ($qtda == 0)) { $qtdf = $qtda = 1; }
			desloca($qtdf,$qtda,\@lenf,\@lena,\@listaf,\@listaa,\@matrix);
			$contf = $contf-$qtdf;
			$conta = $conta-$qtda;			
		}
		@alf = @ala = ();
		if ($qtdf > 0) { @alf = splice(@idsf,0,$qtdf); }
		if ($qtda > 0) { @ala = splice(@idsa,0,$qtda); }
		if ($etiquetado && ($alf[0] =~ /[^\.]+\.1.s1/)) { imprime_cabecalho(nome($arqfonte).'.al',$cabf); }
		if ($etiquetado && ($ala[0] =~ /[^\.]+\.1.s1/)) { imprime_cabecalho(nome($arqalvo).'.al',$caba); }
		imprime_alinhamento(nome($arqfonte).'.al',\%sfonte,\@alf,\@ala,$etiquetado,$idali);
		imprime_alinhamento(nome($arqalvo).'.al',\%salvo,\@ala,\@alf,$etiquetado,$idali++);
		print "Alignment $qtdf:$qtda\n";	
# 		if ($qtdf != $qtda) { 
# 			$saida = 0; 
# 			print "VERIFY: $qtdf:$qtda = ",join(" ",@alf),"<=>",join(" ",@ala),"\n";
# 			print join("\n",map($sfonte{$_},@alf)),"\n",join("\n",map($salvo{$_},@ala)),"\n\n";
# 		}
 	}
	while ($#idsf >= 0) { 
		@alf = splice(@idsf,0,1);
		@ala = ();
		imprime_alinhamento(nome($arqfonte).'.al',\%sfonte,\@alf,\@ala,$etiquetado,$idali++);
		print "Alignment 1:0\n";
# 		$saida = 0;
	}
	while ($#idsa >= 0) { 
		@ala = splice(@idsa,0,1);
		@alf = ();
		imprime_alinhamento(nome($arqalvo).'.al',\%salvo,\@ala,\@alf,$etiquetado,$idali++);
		print "Alignment 0:1\n";
# 		$saida = 0;
	}
	if ($etiquetado) {
		imprime_fim(nome($arqfonte).'.al');
		imprime_fim(nome($arqalvo).'.al');
	}
	print "\n";
	return $saida;
}
	
sub le_sentencas_etiquetadas {
	my($arq,$hashsent) = @_;
	
	my($cab,$idpar,$idsent,$sent,$id,@aux);
	
	$idpar = $idsent = 0;
	open(FILE,$arq) or die "It is not possible to open file $arq";
	while (<FILE>) {
		$sent = '';
		if (/<text [^>]+>/) { $cab = $_; }
		else {
			if (/<p>/) { 
				$idpar++; 
				$idsent = 0;
			}
			if (/<s>/) { 
				$_ =~ s/\n//g;
				@aux = split(/<\/s>/,$_);
				while ($#aux >= 0) {
					$sent = shift(@aux);
					$sent =~ s/<[s|p]>//g;
					$idsent++; 
					$id = nome($arq).'.'.$idpar.'.s'.$idsent; 
					$$hashsent{$id} = $sent; 
				}
			}
			elsif (($_ !~ /<\/p>/) && ($_ !~ /<\/text>/)) { 
				$sent = $_; 
				$sent =~ s/\n//g;
				$idsent++; 
				$$hashsent{$idsent} = $sent; 
			}
		}
	}
	close FILE;
	return $cab;
}

sub le_sentencas_cruas {
	my($arq,$hashsent) = @_;
	
	my($idsent);
	
	$idsent = 1;
	open(FILE,$arq) or die "It is not possible to open file $arq";
	while (<FILE>) {
		if ($_ =~ /\w/) {
			$_ =~ s/[\n\r]//g;			
			$$hashsent{$idsent++} = $_;
		}		
	}
	close FILE;	
}


sub desloca {
	my($qtdl,$qtdc,$lenf,$lena,$listaf,$listaa,$matrix) = @_;
	my($i);
	
	for($i = 1;$i <= $qtdl;$i++) { shift(@$lenf); shift(@$listaf); }
	for($i = 1;$i <= $qtdc;$i++) { shift(@$lena); shift(@$listaa); }
	
	while ($qtdl > 0) {
		shift(@$matrix);
		$qtdl--;
	}
	while ($qtdc > 0) {
		for($i=0;$i <= $#$matrix;$i++) { shift(@{$$matrix[$i]}); }
		$qtdc--;
	}
}

###############################################################################################################
#                                                	LPA
###############################################################################################################
# Retorna em $resp todas as ocorrencias de $palavra em $hash: ela mesma, truncada (com *) ou em expressoes.
sub match {
  my($palavra,$hash) = @_;
  my($h,$h1,@tru,@resp);

  $palavra = quotemeta($palavra);
  @resp = grep(/^$palavra$|^$palavra | ($palavra)$/,keys %{$hash});
  @tru = grep(/\*/,keys %{$hash}); #todas as palavras truncadas (com *)
  while ($#tru >= 0) {
	$h = pop(@tru);
	$h1 = substr($h,0,length($h)-1);
	if ($palavra =~ /$h1/) { push(@resp,$h); }
  }  
  return @resp;	
};

# Retira de $aux todas as ocorrencias nao presentes em $sentenca e altera a sentenca removendo as palavras
# ja testadas.
sub expressoes {
	my($sentenca,$aux) = @_;
	my($min,$i,$faz,$sub,$palavra);
	
	if ($#{$aux} >= 0) {
		$min = lc($$sentenca);
		$faz = 0;
		$i = 0;
		$palavra = $$aux[$i];
		while ($i <= $#{$aux}) {
			if (index($min,$$aux[$i]) == -1) { 
			  delete($$aux[$i]);
			} else {
				$faz = 1;
				if (length($palavra) > length($$aux[$i])) {
					$palavra = $$aux[$i];
				}
			}
			$i++;
		}
		if ($faz) {
			$i = index($min,$palavra);
			$sub = substr($$sentenca,0,$i);
			$sub .= substr($$sentenca,$i+length($palavra),length($$sentenca)-length($palavra));
		}
		$$sentenca = $sub;
	}
}

# $lpa = arquivo contendo a lista de palavras ancoras no formato:
# palavra_fonte <> palavra_alvo
# Le a LPA e armazena as palavras fonte e alvo em $htrans e $vtrans
sub le_LPA {  
  my($lpa,$htrans,$vtrans) = @_;
  my($hword,$vword);

  if (defined $lpa) {
    print "Reading LPA...";
    open(T, $lpa) || die "It is not possible to open file $lpa: $!\n";
    while (<T>) {
		chop;
		($hword, $vword) = split(/ <> /, $_);
		push(@{$$htrans{$hword}}, $vword);
		push(@{$$vtrans{$vword}}, $hword);		
    };
    close(T);
    print " OK!\n";
  };
};

###############################################################################################################
#                                                LISTAS DE FEATURES
###############################################################################################################
# Insere em $lista, na posicao $id, as listas de palavras ancoras, candidatas a nome
# proprio e caracteres especiais de $sentenca
sub listas {
    	my($sentenca,$id,$hash,$lista) = @_;
    	my(@aux,@ancoras,@maiusculas,@caracteres,$i,@tokens,%hash,$x);

    	@ancoras = @maiusculas = @caracteres = ();  
	@tokens = separa_tokens($sentenca);
	
	@caracteres = grep(&caracter,@tokens);
	@tokens = grep(&palavra,@tokens); # so palavras
	@maiusculas = grep(substr($_,0,1) =~ /[A-Z]/,@tokens);	
	
	$i = 0;
	while ($i <= $#tokens) {
		@aux = match($tokens[$i],$hash);			
		&expressoes(\$sentenca,\@aux);
		$x = 0;
		while ($x <= $#aux) { 
			if (defined $aux[$x]) {	push(@ancoras,$aux[$x]); }
			$x++;		
		}
		$i++;
	}
			
	if ($#ancoras >= 0) { push(@{$$lista[$id]{'a'}},@ancoras); }
	if ($#caracteres >= 0) { push(@{$$lista[$id]{'c'}},@caracteres); }
	if ($#maiusculas >= 0) { @{$$lista[$id]{'m'}} = @maiusculas;	}
};  

###############################################################################################################
#                                                MATRIZ
###############################################################################################################
# Calcula o indice de correspondencia entre duas sentencas a partir do tamanho delas
# - $tamf e $tama.
sub indice {
	my($tamf,$tama) = @_;
	
	return (abs($tamf-$tama))/(($tamf+$tama)/2);
}

# Retorna o $valor alterado em relacao ao indice $ind e os parametros.
sub sum {
	my($valor,$comb,$ind) = @_;
		
	if ($ind <= $lowind) {	$valor++; } 
	elsif ($ind >= $highind) { $valor--; }
	if (($comb == 1) && ($ind >= $toohigh)) { #or (($comb == 2) and ($ind >= $lowind))) {
		$valor = -10;
	}	
	return $valor;
}

# Retorna o numero de elementos em comum nos vetores $auxf e $auxa.
sub compara_vetores {
	my($auxf,$auxa,$resp) = @_;
	my($i,@aux,$elemento);
	
#	print "Comparando: ",join(",",@$auxf)," e ",join(",",@$auxa),"\n";
	@aux = map(lc($_),@$auxa);
	@$resp = grep(pertence(lc($_),\@aux),@$auxf);
#	if ($#$resp >= 0) { print "Em comun: ",join(" ",@$resp),"\n"; }
}

# Retorna em $anc um hash tendo como chaves todas as palavras ancoras encontradas em
# $ancf e como valores suas correspondentes em $anca.
sub ancoras {
	my($ancf,$anca,$anc,$htrans) = @_;
	my($i,$j,$pfonte,$palvo,@aux,@jatestados);
	
	@aux = @jatestados = ();
	for($i = 0;$i <= $#$ancf;$i++) {
		$pfonte = quotemeta($$ancf[$i]);
		$j = 0;
		while ($j <= $#$anca) {
			@aux = grep(/^$j$/,@jatestados);
			if ($#aux < 0) {
				$palvo = quotemeta($$anca[$j]);				
				@aux = grep(/^$palvo$/,@{$$htrans{$pfonte}});
				if ($#aux >= 0) {
					push(@{$$anc{$pfonte}},$palvo);
					push(@jatestados,$j);
					$j = $#$anca+1;
				}
			}
			$j++;
		}
	}
}

sub zera_matrix {
	my($li,$lf,$ci,$cf,$m) = @_;
	my($contl,$contc);
	
	for($contl = $li; $contl <= $lf; $contl++) {
		for($contc = 0; $contc <= $cf; $contc++) { ${$$m[$contl]}[$contc] = 0; }	    
	}
	if ($li != 0) {
		for($contc = $ci; $contc <= $cf; $contc++) {
			for($contl = 0; $contl <= $lf-($lf-$li+1); $contl++) { ${$$m[$contl]}[$contc] = 0; }  
		}
	}
}

sub completa_matrix {
	my($li,$lf,$ci,$cf,$listax,$listay,$ancoras,$sx,$idx,$sy,$idy,$m) = @_;
	my($contl,$contc,%anc,@mai,@char);
	
	for($contl = $li; $contl <= $lf; $contl++) {
		for($contc = 0; $contc <= $cf; $contc++) { 
			%anc = @mai = @char = ();
			ancoras($$listax[$contl]{'a'},$$listay[$contc]{'a'},\%anc,$ancoras);
			${$$m[$contl]}[$contc] += quant(\%anc);			
			compara_vetores(\@{$$listax[$contl]{'c'}},\@{$$listay[$contc]{'c'}},\@char);	
			compara_vetores(\@{$$listax[$contl]{'m'}},\@{$$listay[$contc]{'m'}},\@mai);			
			${$$m[$contl]}[$contc] += $#char+1 + $#mai+1;
			${$$m[$contl]}[$contc] += cognatos($$sx{$$idx[$contl]},$$sy{$$idy[$contc]},\@{$$listax[$contl]{'a'}},\@mai);
		}
	}

	for($contc = $ci; $contc <= $cf; $contc++) {
		for($contl = 0; $contl < $li; $contl++) { 
			%anc = @mai = @char = ();
			ancoras($$listax[$contl]{'a'},$$listay[$contc]{'a'},\%anc,$ancoras);
			${$$m[$contl]}[$contc] += quant(\%anc);
			compara_vetores($$listax[$contl]{'c'},$$listay[$contc]{'c'},\@char);	
			compara_vetores($$listax[$contl]{'m'},$$listay[$contc]{'m'},\@mai);
			${$$m[$contl]}[$contc] += $#char+1 + $#mai+1;
			${$$m[$contl]}[$contc] += cognatos($$sx{$$idx[$contl]},$$sy{$$idy[$contc]},\@{$$listax[$contl]{'a'}},\@mai);
		}
	}

	
}


###############################################################################################################
#                                                COGNATOS
###############################################################################################################
sub lcs {
    my ($str1, $str2) = @_;
    my (@m, $i, $j);
    
    my (@c1) = split('', $str1);
    my (@c2) = split('', $str2);

   if ($c1[0] eq $c2[0]) { $m[0][0] = 1; }
   else { $m[0][0] = 0; };

   for ($i = 1; $i <= $#c1; $i++) {
       if ($c1[$i] eq $c2[0]) {
          do {
            $m[$i][0] = 1;
            $i++;
          } until ($i > $#c1);
       }
       else { $m[$i][0] = 0; };
   };

   for ($j = 1; $j <= $#c2; $j++) {
       if ($c1[0] eq $c2[$j]) {
          do {
            $m[0][$j] = 1;
            $j++;
          } until ($j > $#c2);
       }
       else { $m[0][$j] = 0; };
   };

   for ($i = 1; $i <= $#c1; $i++) {
       for ($j = 1; $j <= $#c2; $j++) {
           if ($c1[$i] eq $c2[$j]) { $m[$i][$j] = $m[$i-1][$j-1] + 1; }
           else { $m[$i][$j] = $m[$i-1][$j] > $m[$i][$j-1] ? $m[$i-1][$j] : $m[$i][$j-1]; };
       };
   };    

   return $m[$#c1][$#c2];
};

sub lcsr {
   my ($fonte, $alvo) = @_;
   my($lcsub,$max); 

   if ($fonte eq $alvo) { return 1; };
   $max = length($fonte) > length($alvo) ? length($fonte) : length($alvo);
   $lcsub = lcs($fonte, $alvo)/$max;
   return $lcsub;
};

# Coloca em $array todos os bigramas de $palavra
sub bigramas {
	my($palavra,$array) = @_;
	my($i);
	
	$i = 0;
	while (($i+1) < length($palavra)) {
		push(@{$array},substr($palavra,$i,2));
		$i++;		
	}		 
}

# Retorna o valor do coeficiente de Dice calculado para as palavras pfonte e palvo
sub Dice {
	my($pfonte,$palvo) = @_;
	my(@bfonte,@balvo,$bi,$a,@aux);
	
	$a = 0;
	&bigramas($pfonte,\@bfonte);
	&bigramas($palvo,\@balvo);
	if (($#bfonte >= 0) and ($#balvo >= 0)) {
		foreach $bi (@bfonte) {
			@aux = grep(/^$bi$/,@balvo);
			if ($#aux >= 0) {
				$a += $#aux + 1;
			}
		}
		if ($a > 0) {
			return ((2*$a)/($#bfonte+$#balvo+2));						
		}else { return 0; }
	}else { return 0; }
}

sub tem_cognato {
	my($palavra,$array) = @_;

	if ($metrica eq 'lcsr') { return grep(lcsr($palavra,$_) >= $limite,@$array); }
	elsif ($metrica eq 'dice') { return grep(Dice($palavra,$_) >= $limite,@$array); }
	return 0;
}

# Retorna o numero de palavras fonte na sentenca fonte que possuem cognatos no lado alvo. As palavras cognatas nao podem pertencer a $anc ou $mai.
sub cognatos {	
	my($sentf,$senta,$anc,$mai) = @_;
	my(@palavrasf,@palavrasa,@aux);
	
	@palavrasf = separa_tokens($sentf);
	@palavrasf = grep(&palavra,@palavrasf); # so palavras
	@aux = map(lc($_),@$anc);
	@palavrasf = grep(pertence(lc($_),\@aux) == 0,@palavrasf);
	@aux = map(lc($_),@$mai);
	@palavrasf = grep(pertence(lc($_),\@aux) == 0,@palavrasf);

	@palavrasa = separa_tokens($senta);
	@palavrasa = grep(&palavra,@palavrasa); # so palavras

	return grep(tem_cognato($_,\@palavrasa),@palavrasf);	
}

###############################################################################################################
#                                                COMBINACOES
###############################################################################################################
sub combinacoes {
	my($contf,$conta,$tamf,$tama,$m) = @_;
	my(%lista,%testados,$soma,$comb,$max,$melhor,$l,$c,$s);
	
    	$max = 0;
	$melhor = '';  
	%lista = %testados = ();  
	&insere(\%lista,\%testados,1,1,0,''); 	
	
	while ($#{$lista{'i'}} >= 0) {		
		&remove(\%lista,\$l,\$c,\$soma,\$comb);
		push(@{$testados{'i'}},$l);
		push(@{$testados{'j'}},$c);
		push(@{$testados{'soma'}},$soma);
		
		if ($soma > $max) {
			$max = $soma;
			$melhor = $comb;
		}
	#	if ($verbose) { print "\n$l $c $soma \ncombinations: $comb\n"; }
	
		if (($l <= $contf) && (($c+1) <= $conta)) {
			$s = indice($$tamf[$l-1],$$tama[$c-1]+$$tama[$c]);
	#		if ($verbose) { print "Indice $l <=> $c,".($c+1).": $s\n"; }
			$s = sum($soma+$$m[$l-1][$c-1]+$$m[$l-1][$c],1,$s);
			&insere(\%lista,\%testados,$l+1,$c+2,$s,$comb."1:2\n");
		}
		
		if ((($l+1) <= $contf) && ($c <= $conta)) {			
			$s = indice($$tamf[$l-1]+$$tamf[$l],$$tama[$c-1]);
	#		if ($verbose) { print "Indice $l,".($l+1)." <=> $c: $s\n"; }
			$s = sum($soma+$$m[$l-1][$c-1]+$$m[$l][$c-1],1,$s);
			&insere(\%lista,\%testados,$l+2,$c+1,$s,$comb."2:1\n");
		}
		if ($l <= $contf) {						
			&insere(\%lista,\%testados,$l+1,$c,$soma,$comb."1:0\n");
		}
		if ($c <= $conta) {			
			&insere(\%lista,\%testados,$l,$c+1,$soma,$comb."0:1\n");
		}
		if (($l <= $contf) && ($c <= $conta)) {			
			$s = indice($$tamf[$l-1],$$tama[$c-1]);
	#		if ($verbose) { print "Indice $l <=> $c: $s\n"; }
			$s = sum($soma+$$m[$l-1][$c-1],0,$s);
			&insere(\%lista,\%testados,$l+1,$c+1,$s,$comb."1:1\n"); 
		}	
	}
	return $melhor;
}

# Insere na lista de combinacoes os valores de $i, $j, $soma e $comb
sub insere {
	my($lista,$testados,$i,$j,$soma,$comb) = @_;
	my($contl,$contt);	
	
	if ($soma < 0) { return 0; }
	$contl = $contt = 0;
	while ($contl <= $#{$$lista{'i'}}) { 
		if (($$lista{'i'}[$contl] == $i) && ($$lista{'j'}[$contl] == $j) && ($$lista{'soma'}[$contl] > $soma)) {
			last;			
		}
		$contl++;
	}	
	while ($contt <= $#{$$testados{'i'}}) { 
		if (($$testados{'i'}[$contt] == $i) && ($$testados{'j'}[$contt] == $j) && 
			 ($$testados{'soma'}[$contt] > $soma)) {
			last;			
		}
		$contt++;
	}	
	if (($contl > $#{$$lista{'i'}}) && ($contt > $#{$$testados{'i'}})) { 
		push(@{$$lista{'i'}},$i);
		push(@{$$lista{'j'}},$j);
		push(@{$$lista{'soma'}},$soma);
		push(@{$$lista{'comb'}},$comb);							
	}
}

# Remove o elemento no topo da lista de combinacoes e retorna os valores em $i, $j
# $soma e $comb.
sub remove {
	my($lista,$i,$j,$soma,$comb) = @_;
	
	$$i = pop(@{$$lista{'i'}});
	$$j = pop(@{$$lista{'j'}});
	$$soma = pop(@{$$lista{'soma'}});
	$$comb = pop(@{$$lista{'comb'}});	
}

###############################################################################################################
#                                                ENTRADA/SAIDA
###############################################################################################################
# Subrotina separa_tokens
# Entrada: $sentenca (string)
# Saida: Uma string com o conteudo de $sentenca, mas com os espacos separando tokens
# Funcao: Insere espacos entre os tokens da string $sentenca. Tokens sao palavras, numeros, caracteres de
# pontuacao e contracoes do ingles como 's ou 've.
# Exemplo: 
# Entrada: $str = 'Esse artigo trata do problema de 1.2% dos casos.'
# Saida: $str = 'Esse artigo trata do problema de 1.2 % dos casos .'
sub separa_tokens {
    my($sentenca) = @_;
        
    if ($sentenca =~ /^\s+$/) { return ''; }          
    $sentenca =~ s/\<(\/*)[^\>]+\>//g; # Remove etiquetas se o texto estiver etiquetado      
    
    # Inserindo espaco nos lugares adequados para fazer a separacao por ele depois - Inicio
    $sentenca =~ s/([\\\/\"\!\\(\)\{\}\[\]\?\\:\;\\\\@\$\%\&\=\+\-\~\>\*])/ $1 /g;    
    
    # INICIO - Trata . e , de maneira especial    
    $sentenca =~ s/\.\.\./ . . . /g;        
    $sentenca =~ s/([0-9])([\.\,])([0-9])/$1_$2_$3/g;
    $sentenca =~ s/([^_])([\,\.])/$1 $2 /g;
    $sentenca =~ s/_([\.\,])_/$1/g;    
    # FIM - Trata . e ,  de maneira especial
            
    # INICIO - Trata 's presente nos textos em ingles como um unico token  
    $sentenca =~ s/ (['|`])/ $1 /g;
    $sentenca =~ s/^(['|`])/ $1 /g;
    $sentenca =~ s/(['|`]) / $1 /g;
    $sentenca =~ s/(['|`])(s|ve|re|m) / $1$2 /g;
    $sentenca =~ s/n(['|`])t / n$1t /g; # Separa n't em (ex: don't, doesn't, didn't, etc)
    $sentenca =~ s/(ca|wo) n(['|`])t /$1n$2t /g; # Mas nao can't ou won't  
    # FIM - Trata 's presente nos textos em ingles como um unico token  
    
    $sentenca =~ s/\s+/ /g; # Remove o excesso de espacos
    $sentenca =~ s/^\s+//g; # Remove espacos no comeco da sentenca 
    return split(/ /,$sentenca);    
}

sub imprime_cabecalho {
	my($arq,$cab) = @_;
	
	open(ARQ,">>$arq") or die "It is not possible to open file $arq\n";
	print ARQ "$cab";
	close ARQ;
}

sub imprime_fim {
	my($arq) = @_;
	
	open(ARQ,">>$arq") or die "It is not possible to open file $arq\n";
	print ARQ "\n<\/p>\n<\/text>\n";
	close ARQ;
}

sub imprime_alinhamento {
	my($arq,$hash,$fontes,$alvos,$op,$idali) = @_;
	my($corresp,$i,$aux);

	if ($op) {
		$corresp = "''";
		if ($#$alvos > 0) { $corresp = "'".join(" ",@$alvos)."'"; }
		elsif ($#$alvos == 0) { $corresp = join(" ",@$alvos); }
	}
	
	$i = 0;
	open(fout,">>$arq") or die "It is not possible to open file $arq\n";	
	while ($i <= $#$fontes) {
		if ($op) { # imprime com id e corresp
			$$fontes[$i] =~ /[^\.]+\.(\d+)\.s(\d+)/;
			if ($2 == 1) {
				if ($1 != 1) { print fout "\n<\/p>\n"; }
				print fout "<p>"; 
			}
			print fout "<s id=$$fontes[$i] corresp=$corresp>",$$hash{$$fontes[$i++]},"<\/s>";
		}		
		else { # imprime com snum
			print fout "<s snum=",$idali,">",$$hash{$$fontes[$i++]},"<\/s>\n";
		}		
	}	
	close fout;	
}

# Imprime a matriz nXn
sub imprime_matrix {
	my($lenf,$lena,$conta,$contf,$m) = @_;
	my($c,$l);
	
	for($c = 0; $c < $conta; $c++) { print  "\t\t$c"; }
	print "\n";
	for($c = 0; $c < $conta; $c++) { print "\t\t",$$lena[$c]; }
	print "\n";
	for($l = 0; $l < $contf; $l++) {
		print "$l ",$$lenf[$l];
		for($c = 0; $c < $conta; $c++) { print "\t\t$$m[$l][$c]"; }	    
		print "\n";
	}	
}

# Imprime as listas de palavras ancoras, maiusculas e caracteres especiais contidas em $lista
sub imprime_listas {
	my($lista) = @_;
	my($i);
	
	for($i = 0; $i <= $#$lista;$i++) {
		print "$i";
		if ($#{$$lista[$i]{'a'}} >= 0) { print "\nAnchors: ",join(",",@{$$lista[$i]{'a'}}); }
		if ($#{$$lista[$i]{'m'}} >= 0) { print "\nUpper Case:",join(",",@{$$lista[$i]{'m'}}); }
		if ($#{$$lista[$i]{'c'}} >= 0) { print "\nCaracters:",join(",",@{$$lista[$i]{'c'}}); }
		print "\n";
	}	
}

###############################################################################################################
#                                                AUXILIARES
###############################################################################################################
sub caracter {
  if (($_ eq '!') || ($_ eq '?') || ($_ eq '%') || ($_ eq '$') || ($_ eq '@')) { return 1; }
  return 0;
}  

sub palavra {
  if (($_ !~ /\W/) && ($_ !~ /[_||]/)) { return 1;}  
  return 0;
}

#retorna o nome do arquivo sem extensoes nem diretorios
sub nome {
  my($arq) = $_[0];
  my(@aux,$c);
  @aux = split(/\//,$arq);
  $arq = $aux[$#aux];
  if ($arq =~ /\./) { do { $c = chop($arq); } until ($c eq '.'); }   
  return $arq;
}

# Retorna o nmero de elementos em todos os arrays dos values de $hash.
sub quant {
	my($hash) = @_;
	my($i,@array,$qtd);
	
	$qtd = 0;
	while (($i,@array) = each(%$hash)) { $qtd += ($#array+1); }
	return $qtd;
}

sub pertence {
	my($elemento,$array) = @_;

	if ($#$array < 0) { return 0; }
	$elemento = quotemeta($elemento);
	my @aux = grep(/^$elemento$/,@$array);
	if ($#aux >= 0) { return 1; }
	return 0;
}

__END__

=head1 NAME

TCAalign - Sentence aligner based on Translation Corpus Aligner (Hofland, 1996)

=head1 SYNOPSIS

TCAalign [options...]

 Options:
-source|s         source text file (required)
-target|t         target text file (required)
-corpus|c         parallel corpus file
-lpa|l            anchor word list
-one_to_one|o     if you want to force 1:1 alignments
-cog_metric|m     you can choose between lcsr or dice (default=dice)
-cog_threshold|th threshold to cognate metric (default=0.65)
-tagged|tg        if input texts are tagged with sentence boundaries
-verbose|v        verbose
-help|h|?         this guide

Usage Example:

TCAalign -s pt.txt -t es.txt -l lpa.txt -m lcsr -l 0.70

Helena de Medeiros Caseli 2005

=cut
